/*
 * DOOM v1.9 to v1.91 patcher
 *
 * Should compile with Borland C++ 3.1 at least
 * I can't code so excuse me something below seems really
 * stupid. :P
 */

#include <io.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/***************************************************************************/

typedef struct
  {
    const unsigned long int offset;
    const unsigned char length;
    const unsigned char *olddata;
    const unsigned char *newdata;
  }
patch_t;

/***************************************************************************/

#define NUMEXES 3
#define DOOM_NUMPATCHES 18
#define UDOOM_NUMPATCHES 19
#define FDOOM_NUMPATCHES 21

typedef enum
  {
    DOOM,
    UltimateDOOM,
    FinalDOOM,
    none
  }
doomver_t;

unsigned long exesizetab[NUMEXES] =
  {
    709905,
    715493,
    722629
  };

const char *exedesctab[NUMEXES] =
  {
    "DOOM\" or \"DOOM 2: Hell on Earth",
    "Ultimate DOOM",
    "Final DOOM"
  };

unsigned char exepatchcounttab[NUMEXES] =
  {
    DOOM_NUMPATCHES,
    UDOOM_NUMPATCHES,
    FDOOM_NUMPATCHES
  };

/*
 * Note that the last patch should always be the one that will disable
 * the modified game warning completely
 */
patch_t DOOM_Patches[DOOM_NUMPATCHES] =
  {
    /* Startup banner minor version */
    { 0x60026, 1, "\x09", "\x5B" },       /* DOOM mode */
    { 0x60031, 1, "\x09", "\x5B" },       /* DOOM 2 mode */

    /* Force modified game warning always on */
    { 0x6056A, 2, "\x74\x4C", "\x90\x90" },

    /* Record code modifications by cph */
    { 0x62653, 3, "\xC1\xE0\x08", "\x8A\x22\x42" },
    { 0x638BB, 3, "\xC1\xE2\x08", "\x8A\x33\x43" },
    { 0x63912, 5, "\x05\x80\x00\x00\x00", "\x43\x88\x03\x90\x90" },
    { 0x63918, 3, "\xC1\xF8\x08", "\x90\x90\x90" },
    { 0x6391C, 1, "\x03", "\x23" },
    { 0x6392A, 1, "\x03", "\x04" },
    { 0x63965, 3, "\xC1\xE0\x08", "\x8A\x23\x43" },
    { 0x63A4D, 1, "\x6D", "\x6F" },
    { 0x63B19, 1, "\x6D", "\x6F" },

    /* Modified internal demo lump names ("DEMOxLT") */
    { 0xA9455, 2, "\x00\x00", "lt" },
    { 0xA9465, 2, "\x00\x00", "lt" },
    { 0xA9475, 2, "\x00\x00", "lt" },

    /* Modified warning banner */
    { 0xA9A5B, 251, " This version of DOOM has been modified.  If you would like to\n"\
                    "get a copy of the original game, call 1-800-IDGAMES or see the readme file.\n"\
                    "        You will not receive technical support for modified games.\n"\
                    "                      press enter to continue",

                    "This version of DOOM has been modified. It records demos using\n"\
                    "higher precision than the original version did. These demos are NOT\n"\
                    "compatible with the original engine. See the documentation on how to get\n"\
                    "rid of this notice. PRESS ENTER TO CONTINUE... " },

    /* Modified demo version error msg */
    { 0xAA3B0, 23, "from a different game v", "not from version 1.91.\x00" },

    /* Disable modified game warning completely */
    { 0x6056A, 2, "\x90\x90", "\xEB\x4C" },
  };

patch_t UDOOM_Patches[UDOOM_NUMPATCHES] =
  {
    /* Startup banner minor version */
    { 0x60246, 1, "\x09", "\x5B" },       /* Ultimate DOOM mode */
    { 0x60251, 1, "\x09", "\x5B" },       /* DOOM 2 mode */

    /* Force modified game warning always on */
    { 0x6078A, 2, "\x74\x4C", "\x90\x90" },

    /* Record code modifications by cph */
    { 0x62873, 3, "\xC1\xE0\x08", "\x8A\x22\x42" },
    { 0x63AFB, 3, "\xC1\xE2\x08", "\x8A\x33\x43", },
    { 0x63B52, 5, "\x05\x80\x00\x00\x00", "\x43\x88\x03\x90\x90" },
    { 0x63B58, 3, "\xC1\xF8\x08", "\x90\x90\x90" },
    { 0x63B5C, 1, "\x03", "\x23" },
    { 0x63B6A, 1, "\x03", "\x04" },
    { 0x63BA5, 3, "\xC1\xE0\x08", "\x8A\x23\x43" },
    { 0x63C8D, 1, "\x6D", "\x6F" },
    { 0x63D59, 1, "\x6D", "\x6F" },

    /* Modified internal demo lump names ("DEMOxLT") */
    { 0xAA8BD, 2, "\x00\x00", "lt" },
    { 0xAA8CD, 2, "\x00\x00", "lt" },
    { 0xAA8D5, 2, "\x00\x00", "lt" },
    { 0xAA8DD, 2, "\x00\x00", "lt" },

    /* Modified warning banner */
    { 0xAAF5F, 251, " This version of DOOM has been modified.  If you would like to\n"\
                    "get a copy of the original game, call 1-800-IDGAMES or see the readme file.\n"\
                    "        You will not receive technical support for modified games.\n"\
                    "                      press enter to continue",

                    "This version of DOOM has been modified. It records demos using\n"\
                    "higher precision than the original version did. These demos are NOT\n"\
                    "compatible with the original engine. See the documentation on how to get\n"\
                    "rid of this notice. PRESS ENTER TO CONTINUE... " },

    /* Modified demo version error msg */
    { 0xAB8BC, 23, "from a different game v", "not from version 1.91.\x00" },

    /* Disable modified game warning completely */
    { 0x6078A, 2, "\x90\x90", "\xEB\x4C" },
  };

patch_t FDOOM_Patches[FDOOM_NUMPATCHES] =
  {
    /* Startup banner minor version */
    { 0x60B16, 1, "\x09", "\x5B" },       /* Ultimate DOOM mode */
    { 0x60B2A, 1, "\x09", "\x5B" },       /* Plutonia mode */
    { 0x60B3E, 1, "\x09", "\x5B" },       /* TNT mode */
    { 0x60B49, 1, "\x09", "\x5B" },       /* DOOM 2 mode */

    /* Force modified game warning always on */
    { 0x6107A, 2, "\x74\x4C", "\x90\x90" },

    /* Record code modifications by cph */
    { 0x63153, 3, "\xC1\xE0\x08", "\x8A\x22\x42" },
    { 0x643DB, 3, "\xC1\xE2\x08", "\x8A\x33\x43", },
    { 0x64432, 5, "\x05\x80\x00\x00\x00", "\x43\x88\x03\x90\x90" },
    { 0x64438, 3, "\xC1\xF8\x08", "\x90\x90\x90" },
    { 0x6443C, 1, "\x03", "\x23" },
    { 0x6444A, 1, "\x03", "\x04" },
    { 0x64485, 3, "\xC1\xE0\x08", "\x8A\x23\x43" },
    { 0x6456D, 1, "\x6D", "\x6F" },
    { 0x64639, 1, "\x6D", "\x6F" },

    /* Modified internal demo lump names ("DEMOxLT") */
    { 0xABE1D, 2, "\x00\x00", "lt" },
    { 0xABE2D, 2, "\x00\x00", "lt" },
    { 0xABE35, 2, "\x00\x00", "lt" },
    { 0xABE3D, 2, "\x00\x00", "lt" },

    /* Modified warning banner */
    { 0xAC5A7, 251, " This version of DOOM has been modified.  If you would like to\n"\
                    "get a copy of the original game, call 1-800-IDGAMES or see the readme file.\n"\
                    "        You will not receive technical support for modified games.\n"\
                    "                      press enter to continue",

                    "This version of DOOM has been modified. It records demos using\n"\
                    "higher precision than the original version did. These demos are NOT\n"\
                    "compatible with the original engine. See the documentation on how to get\n"\
                    "rid of this notice. PRESS ENTER TO CONTINUE... " },

    /* Modified demo version error msg */
    { 0xACF04, 23, "from a different game v", "not from version 1.91.\x00" },

    /* Disable modified game warning completely */
    { 0x6107A, 2, "\x90\x90", "\xEB\x4C" },
  };

patch_t *exepatchtab[NUMEXES] =
  {
    DOOM_Patches,
    UDOOM_Patches,
    FDOOM_Patches
  };

/***************************************************************************/

unsigned long filesize (FILE *f)
  {
    unsigned long i = ftell(f), l;
    fseek (f, 0, SEEK_END);
    l = ftell(f);
    fseek (f, i, SEEK_SET);
    return l;
  }

unsigned char errstr[80];
void error (unsigned char errcode)
  {
    fprintf (stderr, errstr);
    exit (errcode);
  }

int argcount; char **args;
unsigned char checkparm (char *p)
  {
    unsigned char i;
    for (i=0; i<argcount; i++)
      if (!strcmp(args[i],p))
        return i;

    return 0;
  }

void showhelp (void)
  {
    sprintf(errstr,
      "  Usage: doomp191 [options] <doomexe>\r\n"\
      "\r\n"\
      "  Options:\r\n"
      "\r\n"\
      "   -h/-?...: Display this help message.\r\n"\
      "   -o......: Don't make a backup copy of <doomexe> first.\r\n"\
      "\r\n"\
      "  Please read the included documentation for important notes\r\n"\
      "  on demo compatibility!\r\n");
    error(1);
  }

void verifyolddata (FILE *f, patch_t patch)
  {
    unsigned char *buf;
    buf = (unsigned char *)malloc(patch.length);
    memset (buf, 0, patch.length);
    if (buf==NULL)
      {
        sprintf(errstr, "  Couldn't allocate %u bytes of memory.\r\n",
         patch.length);
        error(5);
      }
    fseek(f, patch.offset, SEEK_SET);
    fread(buf, patch.length, 1, f);
    if (memcmp(buf, patch.olddata, patch.length))
      {
        sprintf(errstr, "  Invalid old data found at offset 0x%05lX.\r\n",
                 patch.offset);
        error(4);
      }
    free(buf);
    return;
  }

void applypatch (FILE *f, patch_t patch)
  {
    fseek(f, patch.offset, SEEK_SET);
    fwrite(patch.newdata, patch.length, 1, f);
    printf("  Wrote %u byte%s to offset 0x%05lX.\r\n",
     patch.length, patch.length==1 ? "" : "s", patch.offset);
    return;
  }

int file_exists(char *name)
  {
    return (access(name, 0) == 0);
  }

/***************************************************************************/

/*
 * 0 = no error
 * 1 = user error :-)
 * 2 = cannot open <doomexe>
 */
int main (int argc, char **argv)
  {
    unsigned char makebackup, removenoticemode;

    char *exename;
    FILE *exe;
    unsigned long exesize;
    doomver_t exever;
    int i;
    patch_t *patchset;

    printf ("\r\n"\
            " DOOM v1.9 to v1.91 patcher, version 1.0 [%s %s]\r\n"\
            " Assembly hacks by Colin Phipps with help and input from Dashiva\r\n"
            " vv, BahdKo and ocelot. Patcher utility written by vv.\r\n"
            "\r\n", __DATE__, __TIME__);

    argcount = argc; args = argv;

    /* help requested or wrong number of parms */
    if (checkparm("-h") || checkparm("-?") || argcount < 2)
      showhelp();

    /* -o = overwrite, -rn = remove startup notice from v1.91 */
    makebackup = checkparm("-o") ? 0 : 1;
    removenoticemode = checkparm("-rn") ? 1 : 0;

    /* find exe name from command line */
    for (i=0; i<argcount; i++)
      if (args[i][0] != '-') exename = args[i];
    if (exename==NULL) showhelp();
    if (strlen(exename) > 12) exename[12] = 0;   /* cut to 8.3+0 */


    exe = fopen(exename, "rb+");
    if (exe==NULL)
      {
        sprintf(errstr,"  Couldn't open file `%s'.\r\n", exename);
        error(2);
      }

    exesize = filesize(exe);

    /* find which one of the supported exes this is, if any */
    exever = none;
    for (i=0; i<NUMEXES; i++)
      {
        if (exesize==exesizetab[i])
          {
            exever=i;
            break;
          }
      }
    if (exever == none)
      {
        sprintf(errstr, "  `%s' doesn't look like a supported version of DOOM (%lu bytes).\r\n", exesize);
        error(3);
      }

    patchset = exepatchtab[exever];

    /* remove the annoying startup notice from an already patched exe */
    if (removenoticemode)
      {
        printf ("  Removing startup notice from `%s'...\r\n", exename);
        verifyolddata (exe, patchset[exepatchcounttab[exever]-1]);
        applypatch (exe, patchset[exepatchcounttab[exever]-1]);
        printf ("  Done!\r\n");
      }
    /* patch from v1.9 to v1.9 */
    else
      {
        if (makebackup)
          {
            char bakname[14], *dot, cmd[40];

            strcpy (bakname, exename);
            dot = strchr(bakname,'.')+1;
            strncpy (dot, "v19", 3);

            if (file_exists(bakname))
              {
                sprintf(errstr, "  Can't make a backup, file `%s' already exists.\r\n",
                 bakname);
                error(7);
              }

            printf ("  Creating a backup copy called `%s'...\r\n", bakname);
            sprintf (cmd, "copy %s %s > nul", exename, bakname);
            system (cmd); /* chuckle */
            if (!file_exists(bakname))
              {
                sprintf(errstr, "  Couldn't create a backup.\r\n");
                error(6);
              }
          }

        printf("  Patching `%s' [%lu bytes, \"%s\"] ...\r\n",
               exename, exesize, exedesctab[exever]);

        /* again, remember that the last patch in each set is for disabling
           the stupid warning message on startup, so we skip it */

        /* 2 passes to make sure the whole file is ok before patching */
        for (i=0; i<exepatchcounttab[exever]-1; i++)
          verifyolddata(exe, patchset[i]);
        for (i=0; i<exepatchcounttab[exever]-1; i++)
          applypatch(exe, patchset[i]);

        printf("  Patching complete. `%s' is now v1.91!\r\n", exename);
      }

    fclose(exe);
    return 0;
  }
